// Copyright 2021-2025 The Khronos Group
// SPDX-License-Identifier: Apache-2.0

#pragma once

// helium
#include "helium/array/ObjectArray.h"

namespace helide {

using ObjectArray = helium::ObjectArray;

} // namespace helide
