// Copyright 2021-2025 The Khronos Group
// SPDX-License-Identifier: Apache-2.0

#pragma once

// helium
#include "helium/array/Array2D.h"

namespace helide {

using Array2DMemoryDescriptor = helium::Array2DMemoryDescriptor;
using Array2D = helium::Array2D;

} // namespace helide
