// Copyright 2021-2025 The Khronos Group
// SPDX-License-Identifier: Apache-2.0

#pragma once

// helium
#include "helium/array/Array1D.h"

namespace helide {

using Array1DMemoryDescriptor = helium::Array1DMemoryDescriptor;
using Array1D = helium::Array1D;

} // namespace helide
