// Copyright 2021-2025 The Khronos Group
// SPDX-License-Identifier: Apache-2.0

#pragma once

// helium
#include "helium/array/Array3D.h"

namespace helide {

using Array3DMemoryDescriptor = helium::Array3DMemoryDescriptor;
using Array3D = helium::Array3D;

} // namespace helide
