// Copyright 2021-2025 The Khronos Group
// SPDX-License-Identifier: Apache-2.0

#pragma once

#include <cstdint>

namespace helium {

using TimeStamp = uint64_t;
TimeStamp newTimeStamp();

} // namespace helium